import os

import jax
import jax.numpy as jnp
import numpy as np
from flax.training import train_state
from functools import partial

import optax
from .model import VAE, ModelData
from .plotting import plot_training_metrics, plot_reconstructions
from .config import Config
from .io import save_model
import time
from typing import List
from .loss import vae_loss, compute_metrics, TrainValMetrics, cyclical_annealing_beta

__all__ = ['train_vae']



# ------------------------------
# Loss function and training step.
# ------------------------------



@partial(jax.jit, static_argnames=("model",))
def train_step(state, x, rng, model, beta):
    loss, grads = jax.value_and_grad(lambda params: vae_loss(params, x, rng, model, beta).loss)(state.params)
    state = state.apply_gradients(grads=grads)
    return state, loss



# ------------------------------
# Create training state with configurable latent_dim.
# ------------------------------
def create_train_state(rng: jax.random.PRNGKey, latent_dim: int, data_len: int, learning_rate: float):
    model = VAE(latent_dim)
    # Initialize the model with dummy data of shape (1, DATA_LEN)
    params = model.init(rng, jnp.ones((1, data_len)), rng)['params']
    tx = optax.adam(learning_rate)
    state = train_state.TrainState.create(apply_fn=model.apply, params=params, tx=tx)
    return state, model


# ------------------------------
# Training loop.
# ------------------------------
def train_vae(
        train_data,
        val_data,
        config: Config,
        save_dir='vae_outdir',
        print_every=100,
        plot_every=1000,
):
    os.makedirs(save_dir, exist_ok=True)
    data_len = train_data.shape[1]
    t0 = time.time()

    rng = jax.random.PRNGKey(0)
    state, model = create_train_state(rng, config.latent_dim, data_len, config.learning_rate)
    metrics: List[TrainValMetrics] = []
    n_train = train_data.shape[0]
    step = 0
    beta = cyclical_annealing_beta(n_epoch=config.epochs, n_cycle=config.cyclical_annealing_cycles)

    for epoch in range(config.epochs):
        # Shuffle training data indices.
        perm = np.random.permutation(n_train)
        rng, subkey = jax.random.split(rng)
        for i in range(0, n_train, config.batch_size):
            # beta = compute_beta(step, config.cycle_length)
            batch = train_data[perm[i:i + config.batch_size]]
            state, batch_loss = train_step(state, batch, subkey, model, beta[epoch])
            step += 1

        metrics.append(compute_metrics(state, train_data, subkey, model, val_data, beta[epoch]))

        if epoch % print_every == 0:
            print(f"Epoch {epoch}: {metrics[epoch]}")

        if epoch % plot_every == 0:
            plot_training_metrics(metrics, fname=f"{save_dir}/loss.png")
            model_data = ModelData(params=state.params, latent_dim=config.latent_dim)
            plot_reconstructions(model_data, val_data, fname=f"{save_dir}/reconstructions_E{epoch}.png")

    model_data = ModelData(params=state.params, latent_dim=config.latent_dim)
    plot_training_metrics(metrics, fname=f"{save_dir}/loss.png")
    plot_reconstructions(model_data, val_data, fname=f"{save_dir}/reconstructions.png")
    save_model(state, config, metrics, savedir=save_dir)
    # _save_losses(train_losses, val_losses, fname=f"{save_dir}/losses.txt")
    print(f"Training complete. (time: {time.time() - t0:.2f}s)")
